"""
This is the BAGUETTE framework. BAGUETTE stands for Behavioral Analysis Graph Using Execution Traces Towards Explanability.
BAGUETTE is a heterogeneous graph data structure used to represent the behavior of malware samples.
"""